@@ -14,45 +14,59 @@
 //! More specifically, `python3-dll-a` requires `llvm-dlltool` executable
 //! to be present in `PATH` when targeting `*-pc-windows-msvc`.
 //!
-//! Example `build.rs` script
-//! -------------------------
+//! PyO3 integration
+//! ----------------
 //!
-//! The following script can be used to cross-compile Stable ABI
-//! PyO3 extension modules for Windows (64-bit MinGW-w64):
+//! Since version **0.16.4**, the `pyo3` crate implements support
+//! for the Stable ABI Python DLL import library generation via
+//! its new `generate-abi3-import-lib` feature.
 //!
-//! ```no_run
-//! fn main() {
-//!     if std::env::var("TARGET").unwrap() == "x86_64-pc-windows-gnu" {
-//!         let libdir = std::env::var("PYO3_CROSS_LIB_DIR")
-//!             .expect("PYO3_CROSS_LIB_DIR is not set when cross-compiling");
-//!         python3_dll_a::generate_implib(&libdir)
-//!             .expect("python3.dll import library generator failed");
-//!     }
-//! }
+//! In this configuration, `python3-dll-a` becomes a `pyo3` crate dependency
+//! and is automatically invoked by its build script in both native
+//! and cross compilation scenarios.
+//!
+//! ### Example `Cargo.toml` usage for a PyO3 extension module
+//!
+//! ```toml
+//! [dependencies]
+//! pyo3 = { version = "0.16.4", features = ["extension-module", "abi3-py37", "generate-abi3-import-lib"] }
 //! ```
 //!
-//! A compatible `python3.dll` import library will be automatically created in
-//! the directory pointed by `PYO3_CROSS_LIB_DIR` environment variable.
+//! Standalone build script usage
+//! -----------------------------
 //!
-//! If both 64-bit and 32-bit or GNU and MSVC ABI (cross-)compile target
-//! support is needed, the more generic `generate_implib_for_target()`
-//! function must be used:
+//! If an older `pyo3` crate version is used, or a different Python bindings
+//! library is required, `python3-dll-a` can be used directly
+//! from the crate build script.
+//!
+//! The examples below assume using an older version of PyO3.
+//!
+//! ### Example `build.rs` script
+//!
+//! The following cargo build script can be used to cross-compile Stable ABI
+//! PyO3 extension modules for Windows (64/32-bit x86 or 64-bit ARM)
+//! using either MinGW-w64 or MSVC target environment ABI:
 //!
 //! ```no_run
 //! fn main() {
 //!     if std::env::var("CARGO_CFG_TARGET_OS").unwrap() == "windows" {
-//!         let libdir = std::env::var("PYO3_CROSS_LIB_DIR")
+//!         let cross_lib_dir = std::env::var_os("PYO3_CROSS_LIB_DIR")
 //!             .expect("PYO3_CROSS_LIB_DIR is not set when cross-compiling");
 //!         let arch = std::env::var("CARGO_CFG_TARGET_ARCH").unwrap();
 //!         let env = std::env::var("CARGO_CFG_TARGET_ENV").unwrap();
-//!         python3_dll_a::generate_implib_for_target(&libdir, &arch, &env)
+//!
+//!         let libdir = std::path::Path::new(&cross_lib_dir);
+//!         python3_dll_a::generate_implib_for_target(libdir, &arch, &env)
 //!             .expect("python3.dll import library generator failed");
 //!     }
 //! }
 //! ```
 //!
-//! Example `cargo build` invocation
-//! --------------------------------
+//! A compatible `python3.dll` import library file named `python3.dll.a`
+//! or `python3.lib` will be automatically created in the directory
+//! pointed by the `PYO3_CROSS_LIB_DIR` environment variable.
+//!
+//! ### Example `cargo build` invocation
 //!
 //! ```sh
 //! PYO3_CROSS_LIB_DIR=target/python3-dll cargo build --target x86_64-pc-windows-gnu
@@ -64,7 +78,7 @@
 use std::fs::create_dir_all;
 use std::fs::File;
 use std::io::{BufWriter, Error, ErrorKind, Result, Write};
-use std::path::{Path, PathBuf};
+use std::path::Path;
 use std::process::Command;
 
 /// Stable ABI Python DLL file name
@@ -88,6 +102,9 @@ const DLLTOOL_GNU_32: &str = "i686-w64-mingw32-dlltool";
 /// Canonical `dlltool` program name for the MSVC environment ABI (LLVM dlltool)
 const DLLTOOL_MSVC: &str = "llvm-dlltool";
 
+/// Canonical `lib` program name for the MSVC environment ABI (MSVC lib.exe)
+const LIB_MSVC: &str = "lib.exe";
+
 /// Python Stable ABI symbol defs from the CPython repository
 ///
 /// Upstream source: <https://github.com/python/cpython/blob/main/Misc/stable_abi.txt>
@@ -96,7 +113,7 @@ const STABLE_ABI_DEFS: &str = include_str!("../Misc/stable_abi.txt");
 /// Generates `python3.dll` import library directly from the embedded
 /// Python Stable ABI definitions data for the specified compile target.
 ///
-/// The import library file named `python3.dll.a` is created
+/// The import library file named `python3.dll.a` or `python3.lib` is created
 /// in directory `out_dir`.
 ///
 /// The compile target architecture name (as in `CARGO_CFG_TARGET_ARCH`)
@@ -104,10 +121,10 @@ const STABLE_ABI_DEFS: &str = include_str!("../Misc/stable_abi.txt");
 ///
 /// The compile target environment ABI name (as in `CARGO_CFG_TARGET_ENV`)
 /// is passed in `env`.
-pub fn generate_implib_for_target(out_dir: &str, arch: &str, env: &str) -> Result<()> {
+pub fn generate_implib_for_target(out_dir: &Path, arch: &str, env: &str) -> Result<()> {
     create_dir_all(out_dir)?;
 
-    let mut defpath = PathBuf::from(out_dir);
+    let mut defpath = out_dir.to_owned();
     defpath.push(DEF_FILE);
 
     let stable_abi_exports = parse_stable_abi_defs(STABLE_ABI_DEFS);
@@ -117,13 +134,19 @@ pub fn generate_implib_for_target(out_dir: &str, arch: &str, env: &str) -> Resul
     drop(writer);
 
     // Try to guess the `dlltool` executable name from the target triple.
-    let dlltool = match (arch, env) {
+    let (command, dlltool) = match (arch, env) {
         // 64-bit MinGW-w64 (aka x86_64-pc-windows-gnu)
-        ("x86_64", "gnu") => DLLTOOL_GNU,
+        ("x86_64", "gnu") => (Command::new(DLLTOOL_GNU), DLLTOOL_GNU),
         // 32-bit MinGW-w64 (aka i686-pc-windows-gnu)
-        ("x86", "gnu") => DLLTOOL_GNU_32,
+        ("x86", "gnu") => (Command::new(DLLTOOL_GNU_32), DLLTOOL_GNU_32),
         // MSVC ABI (multiarch)
-        (_, "msvc") => DLLTOOL_MSVC,
+        (_, "msvc") => {
+            if let Some(command) = find_lib_exe(arch) {
+                (command, LIB_MSVC)
+            } else {
+                (Command::new(DLLTOOL_MSVC), DLLTOOL_MSVC)
+            }
+        }
         _ => {
             let msg = format!("Unsupported target arch '{arch}' or env ABI '{env}'");
             return Err(Error::new(ErrorKind::Other, msg));
@@ -131,7 +154,7 @@ pub fn generate_implib_for_target(out_dir: &str, arch: &str, env: &str) -> Resul
     };
 
     // Run the selected `dlltool` executable to generate the import library.
-    let status = build_dlltool_command(dlltool, arch, &defpath, out_dir).status()?;
+    let status = build_dlltool_command(command, dlltool, arch, &defpath, out_dir).status()?;
 
     if status.success() {
         Ok(())
@@ -141,12 +164,34 @@ pub fn generate_implib_for_target(out_dir: &str, arch: &str, env: &str) -> Resul
     }
 }
 
+/// Find Visual Studio lib.exe on Windows
+#[cfg(windows)]
+fn find_lib_exe(arch: &str) -> Option<Command> {
+    let target = match arch {
+        "x86_64" => "x86_64-pc-windows-msvc",
+        "x86" => "i686-pc-windows-msvc",
+        "aarch64" => "aarch64-pc-windows-msvc",
+        _ => return None,
+    };
+    cc::windows_registry::find(target, LIB_MSVC)
+}
+
+#[cfg(not(windows))]
+fn find_lib_exe(_arch: &str) -> Option<Command> {
+    None
+}
+
 /// Generates the complete `dlltool` executable invocation command.
 ///
-/// Supports both LLVM and MinGW `dlltool` flavors.
-fn build_dlltool_command(dlltool: &str, arch: &str, defpath: &Path, out_dir: &str) -> Command {
-    let mut libpath = PathBuf::from(out_dir);
-    let mut command = Command::new(dlltool);
+/// Supports Visual Studio `lib.exe`, LLVM and MinGW `dlltool` flavors.
+fn build_dlltool_command(
+    mut command: Command,
+    dlltool: &str,
+    arch: &str,
+    defpath: &Path,
+    out_dir: &Path,
+) -> Command {
+    let mut libpath = out_dir.to_owned();
 
     // Check whether we are using LLVM `dlltool` or MinGW `dlltool`.
     if dlltool == DLLTOOL_MSVC {
@@ -167,6 +212,20 @@ fn build_dlltool_command(dlltool: &str, arch: &str, defpath: &Path, out_dir: &st
             .arg(defpath)
             .arg("-l")
             .arg(libpath);
+    } else if dlltool == LIB_MSVC {
+        libpath.push(IMPLIB_FILE_MSVC);
+
+        // lib.exe use their own target architecure names...
+        let machine = match arch {
+            "x86_64" => "X64",
+            "x86" => "X86",
+            "aarch64" => "ARM64",
+            _ => arch,
+        };
+        command
+            .arg(format!("/MACHINE:{}", machine))
+            .arg(format!("/DEF:{}", defpath.display()))
+            .arg(format!("/OUT:{}", libpath.display()));
     } else {
         libpath.push(IMPLIB_FILE_GNU);
 
@@ -180,19 +239,6 @@ fn build_dlltool_command(dlltool: &str, arch: &str, defpath: &Path, out_dir: &st
     command
 }
 
-/// Generates `python3.dll` import library directly from the embedded
-/// Python Stable ABI definitions data for the default 64-bit MinGW-w64
-/// compile target.
-///
-/// The import library file named `python3.dll.a` is created
-/// in directory `out_dir`.
-///
-/// The import library is generated for the default `x86_64-pc-windows-gnu`
-/// cross-compile target.
-pub fn generate_implib(out_dir: &str) -> Result<()> {
-    generate_implib_for_target(out_dir, "x86_64", "gnu")
-}
-
 /// Exported DLL symbol definition
 struct DllExport {
     /// Export symbol name
@@ -249,8 +295,11 @@ fn write_export_defs(writer: &mut impl Write, dll_name: &str, exports: &[DllExpo
 
 #[cfg(test)]
 mod tests {
+    use std::path::PathBuf;
+
     use super::*;
 
+    #[cfg(unix)]
     #[test]
     fn generate() {
         // FIXME: Use "target/<arch>" dirs for temporary files.
@@ -259,10 +308,10 @@ mod tests {
         dir.push("x86_64-pc-windows-gnu");
         dir.push("python3-dll");
 
-        let out_dir = dir.to_str().unwrap();
-        generate_implib(out_dir).unwrap();
+        generate_implib_for_target(&dir, "x86_64", "gnu").unwrap();
     }
 
+    #[cfg(unix)]
     #[test]
     fn generate_gnu32() {
         let mut dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
@@ -270,8 +319,7 @@ mod tests {
         dir.push("i686-pc-windows-gnu");
         dir.push("python3-dll");
 
-        let out_dir = dir.to_str().unwrap();
-        generate_implib_for_target(out_dir, "x86", "gnu").unwrap();
+        generate_implib_for_target(&dir, "x86", "gnu").unwrap();
     }
 
     #[test]
@@ -281,8 +329,7 @@ mod tests {
         dir.push("x86_64-pc-windows-msvc");
         dir.push("python3-dll");
 
-        let out_dir = dir.to_str().unwrap();
-        generate_implib_for_target(out_dir, "x86_64", "msvc").unwrap();
+        generate_implib_for_target(&dir, "x86_64", "msvc").unwrap();
     }
 
     #[test]
@@ -292,8 +339,7 @@ mod tests {
         dir.push("i686-pc-windows-msvc");
         dir.push("python3-dll");
 
-        let out_dir = dir.to_str().unwrap();
-        generate_implib_for_target(out_dir, "x86", "msvc").unwrap();
+        generate_implib_for_target(&dir, "x86", "msvc").unwrap();
     }
 
     #[test]
@@ -303,13 +349,7 @@ mod tests {
         dir.push("aarch64-pc-windows-msvc");
         dir.push("python3-dll");
 
-        let out_dir = dir.to_str().unwrap();
-        generate_implib_for_target(out_dir, "aarch64", "msvc").unwrap();
-    }
-
-    #[test]
-    fn abi_defs_len() {
-        assert_eq!(STABLE_ABI_DEFS.len(), 48836);
+        generate_implib_for_target(&dir, "aarch64", "msvc").unwrap();
     }
 
     #[test]
